# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import re
import ctypes
from .model import Model

TYPE_RE = re.compile(r"([a-zA-Z ]+)\s*(\*?)\s*((?:[0-9]+)?)")

class NativeModel(Model):

    def __init__(self, library, function, mode=ctypes.CDLL):
        super(NativeModel, self).__init__(self._evaluate)
        self._cdll = mode(library)
        self._func = getattr(self._cdll, function)

    def _str_to_type(self, type, is_pointer=False):
        if type == "bool" or type == "_Bool":
            return ctypes.c_bool
        elif type == "char":
            if is_pointer:
                return ctypes.c_char_p
            else:
                return ctypes.c_char
        elif type == "wchar_t":
            if is_pointer:
                return ctypes.c_wchar
            else:
                return ctypes.c_wchar_p
        elif type == "unsigned char":
            return ctypes.c_ubyte
        elif type == "short":
            return ctypes.c_short
        elif type == "unsigned short":
            return ctypes.c_ushort
        elif type == "int":
            return ctypes.c_int
        elif type == "unsigned int":
            return ctypes.c_uint
        elif type == "long":
            return ctypes.c_long
        elif type == "unsigned long":
            return ctypes.c_ulong
        elif type == "__int64" or type == "long long":
            return ctypes.c_longlong
        elif type == "unsigned __int64" or type == "unsigned long long":
            return ctypes.c_ulonglong
        elif type == "float":
            return ctypes.c_float
        elif type == "double":
            return ctypes.c_double
        elif type == "long double":
            return ctypes.c_longdouble
        elif type == "void" and is_pointer:
            return ctypes.c_void_p
        else:
            raise ValueError("unknown type " + str(type))

    def _cast(self, argument, value=None, length=None):
        if hasattr(argument, "type"):
            type = getattr(argument, "type")

            if isinstance(type, str):
                match = TYPE_RE.match(type)

                if match:
                    base_type = match.group(1)
                    is_pointer = match.group(2) == "*"

                    if length is None:
                        if len(match.group(3)) > 0:
                            length = int(match.group(3))
                        elif value is not None and isinstance(value, (list, tuple)):
                            length = len(value)
                        else:
                            length = 1

                    type = self._str_to_type(base_type, is_pointer)

                    if is_pointer and type != ctypes.c_void_p and type != ctypes.c_char_p and type != ctypes.c_wchar_p:
                        type = type * length

                    if value is None:
                        return type()
                    elif isinstance(value, (list, tuple)):
                        return type(*value)
                    else:
                        return type(value)
                else:
                    raise ValueError("unparseable type attribute")
            else:
                if value is None:
                    return type()
                elif isinstance(value, (list, tuple)):
                    return type(*value)
                else:
                    return type(value)
        else:
            raise ValueError("missing type attribute")

    def _evaluate(self, **kwargs):
        nargs = len(self.parameters) + sum([1 if hasattr(r, "asarg") and getattr(r, "asarg") else 0 for r in self.responses])
        args = [None]*nargs
        arg_map = {}
        locals = {}
        length_args = {}

        # first pass aggregates the length arguments
        for parameter in self.parameters:
            if hasattr(parameter, "len_arg"):
                if getattr(parameter, "len_arg") in length_args:
                    if length_args[getattr(parameter, "len_arg")] != len(kwargs[parameter.name]):
                        raise ValueError("arguments using same length argument have different lengths")
                else:
                    length_args[getattr(parameter, "len_arg")] = len(kwargs[parameter.name])

        for k, v in length_args.items():
            kwargs[k] = v

        # second pass places arguments with the order attribute
        for parameter in self.parameters:
            if hasattr(parameter, "order"):
                arg = self._cast(parameter,
                                 value=kwargs[parameter.name],
                                 length=length_args[getattr(parameter, "len_arg")] if hasattr(parameter, "len_arg") else None)
                arg_map[parameter.name] = getattr(parameter, "order")
                args[getattr(parameter, "order")] = arg

        for response in self.responses:
            if hasattr(response, "asarg") and getattr(response, "asarg") and hasattr(response, "order"):
                arg = self._cast(response,
                                 length=length_args[getattr(response, "len_arg")] if hasattr(response, "len_arg") else None)
                arg_map[response.name] = getattr(response, "order")
                locals[response.name] = arg
                args[getattr(response, "order")] = ctypes.byref(arg)

        # third pass fills in any remaining arguments
        index = 0

        for parameter in self.parameters:
            if not hasattr(parameter, "order"):
                while args[index] is not None:
                    index += 1

                arg = self._cast(parameter,
                                 value=kwargs[parameter.name],
                                 length=length_args[getattr(parameter, "len_arg")] if hasattr(parameter, "len_arg") else None)
                arg_map[parameter.name] = index
                args[index] = arg

        for response in self.responses:
            if hasattr(response, "asarg") and getattr(response, "asarg") and not hasattr(response, "order"):
                while args[index] is not None:
                    index += 1

                arg = self._cast(response,
                                 length=length_args[getattr(response, "len_arg")] if hasattr(response, "len_arg") else None)
                arg_map[response.name] = index
                locals[response.name] = arg
                args[index] = ctypes.byref(arg)

        # determine if the function will be returning a value
        leftover = [r for r in self.responses if not hasattr(r, "asarg") or not getattr(r, "asarg")]

        if len(leftover) > 1:
            raise ValueError("native functions can only return one type")
        elif len(leftover) == 1:
            self._func.restype = type(self._cast(leftover[0]))

        # call the function
        value = self._func(*args)

        # process the results
        result = {}

        for response in self.responses:
            if hasattr(response, "asarg") and getattr(response, "asarg"):
                if hasattr(locals[response.name], "__getitem__"):
                    result[response.name] = [x for x in locals[response.name]]
                else:
                    result[response.name] = locals[response.name].value
            else:
                result[response.name] = value

        return result



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import random
import inspect
from .model import Constraint, Lever, Model, Parameter, Response, \
    Uncertainty, CategoricalLever, IntegerLever, PermutationLever, \
    SubsetLever, RealLever, UniformUncertainty, NormalUncertainty, \
    LogNormalUncertainty

MINIMIZE = Response.MINIMIZE
MAXIMIZE = Response.MAXIMIZE
INFO = Response.INFO
IGNORE = Response.IGNORE

class UnnamedObject(object):

    def __init__(self, constructor, *args, **kwargs):
        super(UnnamedObject, self).__init__()
        self.constructor = constructor
        self.args = args
        self.kwargs = kwargs

    def convert(self, name):
        return self.constructor(name, *self.args, **self.kwargs)

class Real(UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Real, self).__init__(RealLever, *args, **kwargs)

class Integer(UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Integer, self).__init__(IntegerLever, *args, **kwargs)

class Categorical(UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Categorical, self).__init__(CategoricalLever, *args, **kwargs)

class Permutation(UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Permutation, self).__init__(PermutationLever, *args, **kwargs)

class Subset(UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Subset, self).__init__(SubsetLever, *args, **kwargs)

class Uniform(float, UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Uniform, self).__init__(UniformUncertainty, *args, **kwargs)

    def __new__(cls, *args, **kwargs):
        return float.__new__(cls, kwargs.get("default_value", float("NaN")))

class Normal(float, UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(Normal, self).__init__(NormalUncertainty, *args, **kwargs)

    def __new__(cls, *args, **kwargs):
        return float.__new__(cls, kwargs.get("default_value", float("NaN")))

class LogNormal(float, UnnamedObject):

    def __init__(self, *args, **kwargs):
        super(LogNormal, self).__init__(LogNormalUncertainty, *args, **kwargs)

    def __new__(cls, *args, **kwargs):
        return float.__new__(cls, kwargs.get("default_value", float("NaN")))


class Minimize(Response):

    def __init__(self, name, **kwargs):
        super(Minimize, self).__init__(name, type=Response.MINIMIZE, **kwargs)

class Maximize(Response):

    def __init__(self, name, **kwargs):
        super(Maximize, self).__init__(name, type=Response.MAXIMIZE, **kwargs)

class Info(Response):

    def __init__(self, name, **kwargs):
        super(Info, self).__init__(name, type=Response.INFO, **kwargs)

class Ignore(Response):

    def __init__(self, **kwargs):
        super(Ignore, self).__init__("Ignored" + str(random.randint()), type=Response.IGNORE, **kwargs)

class CallableModel(Model):

    def __init__(self, function):
        super(CallableModel, self).__init__(function)

    def __call__(self, *args, **kwargs):
        return self.function(*args, **kwargs)

class RhodiumModel(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, f):
        if not hasattr(f, "rhodium_model"):
            f.rhodium_model = CallableModel(f)

        argspec = inspect.getfullargspec(f)
        nargs = len(argspec.args)
        ndefs = len(argspec.defaults)

        for i, k in enumerate(argspec.args):
            if i >= nargs-ndefs:
                # has default value, determine if special rhodium annotation
                def_val = argspec.defaults[i-(nargs-ndefs)]

                if isinstance(def_val, UnnamedObject):
                    named_obj = def_val.convert(k)

                    if isinstance(named_obj, Lever):
                        f.rhodium_model.levers[k] = named_obj
                    elif isinstance(named_obj, Uncertainty):
                        f.rhodium_model.uncertainties[k] = named_obj

            if k not in f.rhodium_model.parameters:
                f.rhodium_model.parameters[k] = Parameter(k)

        return f.rhodium_model

class Parameters(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, f):
        if not hasattr(f, "rhodium_model"):
            f.rhodium_model = CallableModel(f)

        for v in self.args:
            if isinstance(v, str):
                f.rhodium_model.parameters[v] = Parameter(v)
            elif isinstance(v, Parameter):
                f.rhodium_model.parameters[v.name] = v
            else:
                raise ValueError("arg must be a string or Parameter")

        for k, v in self.kwargs.items():
            if isinstance(v, dict):
                f.rhodium_model.parameters[k] = Parameter(k, **v)
            elif isinstance(v, (list, tuple)):
                f.rhodium_model.parameters[k] = Parameter(k, *v)
            else:
                f.rhodium_model.parameters[k] = Parameter(k, v)

        return f

class Responses(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, f):
        if not hasattr(f, "rhodium_model"):
            f.rhodium_model = CallableModel(f)

#         if len(self.args) == 2 and isinstance(self.args[0], (list, tuple)) and isinstance(self.args[1], (list, tuple)):
#             for i, v in enumerate(self.args[0]):
#                 f.rhodium_model.responses[v] = Response(v, self.args[1][i if i < len(self.args[1]) else -1])
#         elif len(self.args) == 1 and isinstance(self.args[0], (list, tuple)):
#             for v in self.args[0]:
#                 f.rhodiuim_model.responses[v] = Response(v, Response.INFO)
#         else:
#             for v in self.args:
#                 if isinstance(v, str):
#                     f.rhodium_model.responses[v] = Response(v, Response.INFO)
#                 elif isinstance(v, Response):
#                     f.rhodium_model.responses[v.name] = v
#                 else:
#                     raise ValueError("arg must be a string or Response")

        for v in self.args:
            if isinstance(v, str):
                f.rhodium_model.responses[v] = Response(v, Response.INFO)
            elif isinstance(v, Response):
                f.rhodium_model.responses[v.name] = v
            else:
                raise ValueError("arg must be a string or Response")

        for k, v in self.kwargs.items():
            if isinstance(v, UnnamedObject):
                f.rhodium_model.responses[k] = v.convert(k)
            elif isinstance(v, dict):
                f.rhodium_model.responses[k] = Response(k, **v)
            elif isinstance(v, (list, tuple)):
                f.rhodium_model.responses[k] = Response(k, *v)
            else:
                f.rhodium_model.responses[k] = Response(k, v)

        return f

class Constraints(object):

    def __init__(self, *args):
        self.args = args

    def __call__(self, f):
        if not hasattr(f, "rhodium_model"):
            f.rhodium_model = CallableModel(f)

        for v in self.args:
            if isinstance(v, Constraint):
                f.rhodium_model.constraints.append(v)
            elif isinstance(v, str) or callable(v):
                f.rhodium_model.constraints.append(Constraint(v))
            else:
                raise ValueError("arg must be a Constraint, string, or callable")

        return f

class Levers(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, f):
        if not hasattr(f, "rhodium_model"):
            f.rhodium_model = CallableModel(f)

        for v in self.args:
            if isinstance(v, Lever):
                f.rhodium_model.levers[v.name] = v
            else:
                raise ValueError("arg must be a Lever")

        for k, v in self.kwargs.items():
            if isinstance(v, UnnamedObject):
                f.rhodium_model.levers[k] = v.convert(k)
            else:
                raise ValueError("kwarg must be a Lever")

        return f

class Uncertainties(object):

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __call__(self, f):
        if not hasattr(f, "rhodium_model"):
            f.rhodium_model = CallableModel(f)

        for v in self.args:
            if isinstance(v, Uncertainty):
                f.rhodium_model.uncertainties[v.name] = v
            else:
                raise ValueError("arg must be an Uncertainty")

        for k, v in self.kwargs.items():
            if isinstance(v, UnnamedObject):
                f.rhodium_model.uncertainties[k] = v.convert(k)
            else:
                raise ValueError("kwarg must be an Uncertainty")

        return f



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import matplotlib as mpl
from collections import OrderedDict
from pandas.core.frame import DataFrame
from .config import RhodiumConfig
from .model import DataSet

class Brush(object):
    """Defines a brush to color data points matching an expression.

    A brush is a mechanism to highlight, or color, data that matches a given
    expression.  The expression is a string that contains one or more boolean
    or comparison operators.  Any parameter in the data set can be referenced.
    For example, if the data set contains parameters "x" and "y", valid
    expressions include:
        x < 5
        y > 10
        x <= 5 and y > 10
    """

    def __init__(self, name, expr=None, color=None):
        super(Brush, self).__init__()

        if expr is None:
            expr = name

        self.name = name
        self.expr = expr
        self.color = color

class BrushSet(object):
    """A collection of Brush objects.

    If two or more brushes overlap, then the first brush, in the order they
    are defiend in this collection, determines the coloring.  For example, for
        BrushSet([Brush("x < 5", "blue"), Brush("x > 3", "green")])
    a point with value x = 4 will be colored "blue".
    """

    def __init__(self, brushes):
        super(BrushSet, self).__init__()
        self.map = OrderedDict()

        if not isinstance(brushes, (list, tuple)):
            brushes = [brushes]

        for brush in brushes:
            if isinstance(brush, Brush):
                self.map[brush.name] = brush
            elif isinstance(brush, BrushSet):
                self.map.update(brush.map)
            elif isinstance(brush, str):
                self.map[brush] = Brush(brush, brush)
            else:
                raise ValueError("only Brush or string expressions can be added to BrushSet")

    def __len__(self):
        return len(self.map)

    def keys(self):
        return self.map.keys()

    def __contains__(self, key):
        return key in self.map

    def __getitem__(self, key):
        return self.map[key]

    def __iter__(self):
        return iter(self.map.values())

    def colors(self):
        return [brush.color for brush in self]

def _as_brush_set(input):
    """Converts the input into a BrushSet.

    The input can be either a single brush or multiple brush, each defined
    as either a string (e.g., "x < 5") or a brush (e.g., Brush("x < 5")).
    """
    if isinstance(input, BrushSet):
        return input
    else:
        return BrushSet(input)

def apply_brush(brush_set, data):
    """Applies the given brush set to the given data, returning the color assignments."""
    if isinstance(data, DataFrame):
        return _apply_brush_dataframe(brush_set, data)
    elif isinstance(data, DataSet):
        return _apply_brush_dataset(brush_set, data)
    else:
        raise ValueError("unsupported type, data must be a DataFrame or DataSet")

def _apply_brush_dataset(brush_set, ds):
    brush_set = _as_brush_set(brush_set)
    n = len(ds)
    assignment = [None]*n

    for brush in brush_set:
        bin = ds.apply(brush.expr)

        for i in range(n):
            if bin[i]:
                assignment[i] = brush.name

    for i, a in enumerate(assignment):
        if a is None:
            assignment[i] = RhodiumConfig.default_unassigned_label

    return assignment

def _apply_brush_dataframe(brush_set, df):
    brush_set = _as_brush_set(brush_set)
    assignment = [None]*df.shape[0]

    for brush in brush_set:
        bin = df.query(brush.expr)

        for i in bin.index:
            if assignment[i] is None:
                assignment[i] = brush.name

    for i, a in enumerate(assignment):
        if a is None:
            assignment[i] = RhodiumConfig.default_unassigned_label

    return assignment

def brush_color_map(brush_set, assignment):
    brush_set = _as_brush_set(brush_set)
    cc = mpl.colors.colorConverter
    color_map = OrderedDict()

    if RhodiumConfig.default_unassigned_label in assignment:
        color_map[RhodiumConfig.default_unassigned_label] = RhodiumConfig.default_unassigned_brush_color

    # we want to retain the original ordering of the brushes, so first determine
    # which brushes are applicable to a data set
    classes = set([a for a in assignment])

    for b in brush_set:
        if b.name in classes:
            color_map[b.name] = cc.to_rgba(b.color) if b.color is not None else None

    # determine if any brushes have no assigned color and pick one
    unassigned_count = sum([1 if x is None else 0 for x in color_map.values()])

    if unassigned_count > 0:
        brush_colors = RhodiumConfig.default_brush_colors
        count = 0

        for k, v in color_map.items():
            if v is None:
                color_map[k] = brush_colors[count]
                count += 1

    return color_map

def color_brush(brush_set, data, **kwargs):
    assignment = apply_brush(brush_set, data)
    color_map = brush_color_map(brush_set, assignment)
    return ([color_map[a] for a in assignment], color_map)

def color_indices(c, color_map):
    values = list(color_map.values())
    return [values.index(v) for v in c]



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys
import unittest
from ..model import Parameter, Response, IntegerUncertainty, \
    UniformUncertainty
from ..optimization import evaluate
from ..sampling import sample_lhs

class TestExcelHelper(unittest.TestCase):

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testGetItem(self):
        from ..excel import ExcelHelper
        file = os.path.join(os.path.dirname(__file__), "TestGetItem.xlsx")

        with ExcelHelper(file) as helper:
            self.assertEqual(1, helper["A1"])
            self.assertEqual(5, helper["A5"])
            self.assertEqual(((1.0,), (2.0,), (3.0,), (4.0,), (5.0,)), helper["A1:A5"])
            self.assertEqual(None, helper["B2"])
            self.assertEqual(u"hello", helper["C2"])
            self.assertEqual(((u"hello", u"world",),), helper["C2:D2"])

            helper.set_sheet(2)
            self.assertEqual(u"sheet 2", helper["B2"])

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testSetItem(self):
        from ..excel import ExcelHelper
        file = os.path.join(os.path.dirname(__file__), "TestSetItem.xlsx")

        with ExcelHelper(file) as helper:
            helper["A1"] = 1
            self.assertEqual(6, helper["A3"])

            helper["A2"] = 3
            self.assertEqual(4, helper["A3"])

            helper["A1:A2"] = [6, 7]
            self.assertEqual(13, helper["A3"])

            helper.set_sheet(2)
            helper["B2"] = "hello"
            helper.set_sheet(1)
            helper["B2"] = "world"
            helper.set_sheet(2)
            self.assertEqual(u"hello", helper["B2"])

class TestExcelModel(unittest.TestCase):

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testEvaluate(self):
        from ..excel import ExcelModel
        file = os.path.join(os.path.dirname(__file__), "TestModel.xlsx")

        with ExcelModel(file) as model:
            model.parameters = [Parameter("X1", cell="B1"),
                                Parameter("X2", cell="B2")]
            model.responses = [Response("Y", cell="B3")]

            result = evaluate(model, {"X1": 3, "X2": 5})
            self.assertEqual(8, result["Y"])

    @unittest.skipUnless(sys.platform.startswith("win"), "requires Windows")
    def testSample(self):
        from ..excel import ExcelModel
        file = os.path.join(os.path.dirname(__file__), "TestModel.xlsx")

        with ExcelModel(file) as model:
            model.parameters = [Parameter("X1", cell="B1"),
                                Parameter("X2", cell="B2")]
            model.responses = [Response("Y", cell="B3")]
            model.uncertainties = [UniformUncertainty("X1", 0.0, 1.0),
                                   IntegerUncertainty("X2", 2, 5)]

            samples = sample_lhs(model, 100)
            results = evaluate(model, samples)

            for result in results:
                self.assertEqual(result["Y"], result["X1"] + result["X2"])



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import ast

_eval_env = {}
_modules = ["math"]

for modulename in _modules:
    module = __import__(modulename, fromlist=[''])
    for name in dir(module):
        if not name.startswith("_"):
            _eval_env[name] = getattr(module, name)

def _has_assignment(tree):
    return isinstance(tree.body[0], ast.Assign)

def _get_result_keys(tree):
    keys = []

    if _has_assignment(tree):
        target = tree.body[0].targets[0]

        if isinstance(target, ast.Tuple):
            keys.extend([t.id for t in target.elts if isinstance(t, ast.Name)])
        elif isinstance(target, ast.Name):
            keys.append(target.id)

    return keys

def _evaluate(expr, env, update_env=True):
    tmp_env = {}
    tmp_env.update(_eval_env)
    tmp_env.update(env)

    if isinstance(expr, str):
        tree = ast.parse(expr, mode="exec")
        result_keys = _get_result_keys(tree)

        if update_env and len(result_keys) > 0:
            exec(expr, {}, tmp_env)

            for key in result_keys:
                env[key] = tmp_env[key]

            if len(result_keys) == 1:
                return tmp_env[result_keys[0]]
            else:
                return [tmp_env[key] for key in result_keys]
        else:
            return eval(expr, {}, tmp_env)
    elif callable(expr):
        tmp_env = {}
        tmp_env.update(env)

        return expr(tmp_env)
    else:
        raise ValueError("expr must be a string or a callable")

def _evaluate_all(expr, iterable, update_env=True):
    result = []

    if isinstance(expr, str):
        tree = ast.parse(expr, mode="exec")
        result_keys = _get_result_keys(tree)

        for env in iterable:
            tmp_env = {}
            tmp_env.update(_eval_env)
            tmp_env.update(env)

            if update_env and len(result_keys) > 0:
                exec(expr, {}, tmp_env)

                for key in result_keys:
                    env[key] = tmp_env[key]

                if len(result_keys) == 1:
                    result.append(tmp_env[result_keys[0]])
                else:
                    result.append([tmp_env[key] for key in result_keys])
            else:
                result.append(eval(expr, {}, tmp_env))
    elif callable(expr):
        for env in iterable:
            tmp_env = {}
            tmp_env.update(env)

            result.append(expr(tmp_env))
    else:
        raise ValueError("expr must be a string or a callable")

    return result



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
from .model import Model

class OpenMDAOModel(Model):

    def __init__(self, problem):
        super(OpenMDAOModel, self).__init__(self._evaluate)

        from openmdao.api import Problem
        if not isinstance(problem, Problem):
            raise ValueError("problem must be an OpenMDAO Problem instance")

        self.problem = problem

    def _evaluate(self, **kwargs):
        result = {}

        for parameter in self.parameters:
            key = parameter.name

            if hasattr(parameter, "connect"):
                key = getattr(parameter, "connect")

            self.problem.root.unknowns[key] = kwargs[parameter.name]

        self.problem.run()

        for response in self.responses:
            key = response.name

            if hasattr(response, "connect"):
                key = getattr(response, "connect")

            result[response.name] = self.problem.root.unknowns[key]

        return result



# Copyright 2015-2024 David Hadka
#
# This file is part of Rhodium, a Python module for robust decision
# making and exploratory modeling.
#
# Rhodium is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Rhodium is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Rhodium.  If not, see <http://www.gnu.org/licenses/>.
import os
import sys

def promptToRun(message, default="yes"):
    if default == "yes":
        prompt = "[Y/n]"
    elif default == "no":
        prompt = "[y/N]"
    else:
        raise ValueError("invalid default answer")

    print(message + " " + prompt + " ", end='')
    sys.stdout.flush()

    if os.getenv("RHODIUM_NO_PROMPT"):
        print("(Skipping prompt since RHODIUM_NO_PROMPT is set)")
        response = ""
    else:
        response = sys.stdin.readline().strip()

    if response == "":
        response = default[0]

    if response == "y" or response == "Y":
        return True
    else:
        return False



import math
import numpy as np
import matplotlib.pyplot as plt
from scipy.optimize import brentq as root
from rhodium import *

# Example using direct policy search (DPS) following the approach of [1]:
#
# [1] Quinn, J. D., P. M. Reed, and K. Keller (2017).  "Direct policy search for
#     robust multi-objective management of deeply uncertain socio-ecological
#     tipping points."  Environmental Modelling & Software, 92:125-141.

# Create a lever for storing the cubic radial basis functions used by this DPS.
# This lever stores one or more cubic radial basis functions defined by a center,
# radius, and weight.  We could have also created a RealLever for each value, but
# creating a class in this manner aids re-usability.
class CubicDPSLever(Lever):

    def __init__(self, name, length = 1, c_bounds = (-2, 2), r_bounds = (0, 2)):
        super(CubicDPSLever, self).__init__(name)
        self.length = length
        self.c_bounds = c_bounds
        self.r_bounds = r_bounds

    # converts from Rhodium levers to Platypus decision variables; a single lever
    # can map to one or more decision variables.
    def to_variables(self):
        result = []

        for _ in range(self.length):
            result += [Real(self.c_bounds[0], self.c_bounds[1])] # the center
            result += [Real(self.r_bounds[0], self.r_bounds[1])] # the radius
            result += [Real(0, 1)]                               # the weight

        return result

    # convert the value of the decision variables from Platypus back into Rhodium;
    # here we create a complex dictionary object storing the radial basis function
    # parameters.
    def from_variables(self, variables):
        policy = {}
        policy["length"] = self.length
        policy["rbfs"] = []

        # extract the parameters for each radial basis function
        for i in range(self.length):
            policy["rbfs"] += [{
                "center" : variables[i*3+0],
                "radius" : variables[i*3+1],
                "weight" : variables[i*3+2] }]

        # normalize the weights
        weight_sum = sum([p["weight"] for p in policy["rbfs"]])

        for i in range(self.length):
            policy["rbfs"][i]["weight"] /= weight_sum

        return policy

# A function for evaluating our cubic DPS.  This is based on equation (12)
# from [1].
def evaluateCubicDPS(policy, current_value):
    value = 0

    for i in range(policy["length"]):
        rbf = policy["rbfs"][i]
        value += rbf["weight"] * abs((current_value - rbf["center"]) / rbf["radius"])**3

    value = min(max(value, 0.01), 0.1)
    return value    

# Construct the lake problem
def lake_problem(policy,  # the DPS policy
         b = 0.42,        # decay rate for P in lake (0.42 = irreversible)
         q = 2.0,         # recycling exponent
         mean = 0.02,     # mean of natural inflows
         stdev = 0.001,   # standard deviation of natural inflows
         alpha = 0.4,     # utility from pollution
         delta = 0.98,    # future utility discount rate
         nsamples = 100,  # monte carlo sampling of natural inflows
         steps = 100):    # the number of time steps (e.g., days)
    Pcrit = root(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    X = np.zeros((steps,))
    decisions = np.zeros((steps,))
    average_daily_P = np.zeros((steps,))
    reliability = 0.0
    utility = 0.0
    inertia = 0.0

    for _ in range(nsamples):
        X[0] = 0.0

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = steps)

        for t in range(1,steps):
            decisions[t-1] = evaluateCubicDPS(policy, X[t-1])
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)

        reliability += np.sum(X < Pcrit)/float(steps)
        utility += np.sum(alpha*decisions*np.power(delta,np.arange(steps)))
        inertia += np.sum(np.diff(decisions) > -0.01)/float(steps-1)

    max_P = np.max(average_daily_P)
    reliability /= float(nsamples)
    utility /= float(nsamples)
    inertia /= float(nsamples)

    return (max_P, utility, inertia, reliability)

model = Model(lake_problem)

model.parameters = [Parameter("policy"),
                    Parameter("b"),
                    Parameter("q"),
                    Parameter("mean"),
                    Parameter("stdev"),
                    Parameter("delta")]

model.responses = [Response("max_P", Response.MINIMIZE),
                   Response("utility", Response.MAXIMIZE),
                   Response("inertia", Response.MAXIMIZE),
                   Response("reliability", Response.MAXIMIZE)]

# Use our new DPS lever
model.levers = [CubicDPSLever("policy", length=3)]

setup_cache(file="example.cache")
output = cache("dps_output", lambda: optimize(model, "NSGAII", 10000))

print(output)

scatter3d(model, output)
plt.show()


import math
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
from scipy.optimize import brentq as root
from rhodium import *

# Construct the lake problem
def lake_problem(pollution_limit,
         b = 0.42,        # decay rate for P in lake (0.42 = irreversible)
         q = 2.0,         # recycling exponent
         mean = 0.02,     # mean of natural inflows
         stdev = 0.001,   # standard deviation of natural inflows
         alpha = 0.4,     # utility from pollution
         delta = 0.98,    # future utility discount rate
         nsamples = 100): # monte carlo sampling of natural inflows
    Pcrit = root(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    nvars = len(pollution_limit)
    X = np.zeros((nvars,))
    average_daily_P = np.zeros((nvars,))
    decisions = np.array(pollution_limit)
    reliability = 0.0

    for _ in range(nsamples):
        X[0] = 0.0

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = nvars)

        for t in range(1,nvars):
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)

        reliability += np.sum(X < Pcrit)/float(nsamples*nvars)

    max_P = np.max(average_daily_P)
    utility = np.sum(alpha*decisions*np.power(delta,np.arange(nvars)))
    inertia = np.sum(np.diff(decisions) > -0.02)/float(nvars-1)

    return (max_P, utility, inertia, reliability)

model = Model(lake_problem)

# Define all parameters to the model that we will be studying
model.parameters = [Parameter("pollution_limit"),
                    Parameter("b"),
                    Parameter("q"),
                    Parameter("mean"),
                    Parameter("stdev"),
                    Parameter("delta")]

# Define the model outputs
model.responses = [Response("max_P", Response.MINIMIZE),
                   Response("utility", Response.MAXIMIZE),
                   Response("inertia", Response.MAXIMIZE),
                   Response("reliability", Response.MAXIMIZE)]

# Define any constraints (can reference any parameter or response by name)
#model.constraints = [Constraint("reliability >= 0.95")]

# Some parameters are levers that we control via our policy
model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

# Some parameters are exogeneous uncertainties, and we want to better
# understand how these uncertainties impact our model and decision making
# process
model.uncertainties = [UniformUncertainty("b", 0.1, 0.45),
                       UniformUncertainty("q", 2.0, 4.5),
                       UniformUncertainty("mean", 0.01, 0.05),
                       UniformUncertainty("stdev", 0.001, 0.005),
                       UniformUncertainty("delta", 0.93, 0.99)]

# Prepare the cache for storing intermediate results
setup_cache(file="example.cache")

# Optimize the model or get cached results if they exist.  Note that the
# call to optimize is wrapped in a lambda function to enable lazy evaluation.
output = cache("output", lambda: optimize(model, "NSGAII", 10000))

# save the Pareto approximate set as a .csv file
output.save('optimization_results.csv')


# ----------------------------------------------------------------------------
# Plotting
# ----------------------------------------------------------------------------

# Use Seaborn settings for pretty plots
sns.set()

# Plot the points in 2D space
scatter2d(model, output)
plt.show()

# The optional interactive flag will show additional details of each point when
# hovering the mouse
scatter2d(model, output, interactive=True)
plt.show()

# Most of Rhodiums's plotting functions accept an optional expr argument for
# classifying or highlighting points meeting some condition
scatter2d(model, output, x="reliability", brush=Brush("reliability >= 0.2"))
plt.show()

# Plot the points in 3D space
scatter3d(model, output, s="reliability", show_legend=True)
plt.show()

# Kernel density estimation plots show density contours for samples.  By
# default, it will show the density of all sampled points
kdeplot(model, output, x="max_P", y="utility")
plt.show()

# Alternatively, we can show the density of all points meeting one or more
# conditions
kdeplot(model, output, x="max_P", y="utility",
        brush=["reliability >= 0.2", "reliability < 0.2"],
        alpha=0.8)
plt.show()

# Pairwise scatter plots shown 2D scatter plots for all outputs
pairs(model, output)
plt.show()

# We can also highlight points meeting one or more conditions
pairs(model, output,
      brush=["reliability >= 0.2", "reliability < 0.2"])
plt.show()

# Joint plots show a single pair of parameters in 2D, their distributions using
# histograms, and the Pearson correlation coefficient
joint(model, output, x="max_P", y="utility")
plt.show()

# A histogram of the distribution of points along each parameter
hist(model, output)
plt.show()

# A parallel coordinates plot to view interactions among responses
parallel_coordinates(model, output, colormap="rainbow", zorder="reliability", brush=Brush("reliability > 0.2"))     
plt.show()

# ----------------------------------------------------------------------------
# Identifying Key Uncertainties
# ----------------------------------------------------------------------------

# The remaining figures look better using Matplotlib's default settings
mpl.rcdefaults()

# We can manually construct policies for analysis.  A policy is simply a Python
# dict storing key-value pairs, one for each lever.
policy = {"pollution_limit" : [0.02]*100}

# Or select one of our optimization results
policy = output[3]

# Construct a specific policy and evaluate it against 1000 states-of-the-world
SOWs = sample_lhs(model, 100)
results = evaluate(model, update(SOWs, policy))

# save the results as a .csv file
results.save('reevaluation_results.csv')

# define performance criteria
metric = ["Reliable" if v["reliability"] > 0.9 else "Unreliable" for v in results]

# Use PRIM to identify the key uncertainties if we require reliability > 0.9
p = Prim(results, metric, include=model.uncertainties.keys(), coi="Reliable")
box = p.find_box()
box.show_details()
plt.show()

# Use CART to identify the key uncertainties
c = Cart(results, metric, include=model.uncertainties.keys())
c.print_tree(coi="Reliable")
c.show_tree()
plt.show()

# Sensitivity analysis using Morris method
print(sa(model, "reliability", policy=policy, method="morris", nsamples=1000, num_levels=4, grid_jump=2))



import math
import bisect
import numbers
import operator
import functools
import numpy as np
import matplotlib.pyplot as plt
from rhodium import * 

##==============================================================================
## Implement the model described by Eijgenraam et al. (2012)
##------------------------------------------------------------------------------

# Parameters pulled from the paper describing each dike ring
params = ("c", "b", "lam", "alpha", "eta", "zeta", "V0", "P0", "max_Pf")
raw_data = {
    10 : ( 16.6939, 0.6258, 0.0014, 0.033027, 0.320, 0.003774,  1564.9, 0.00044, 1/2000),
    11 : ( 42.6200, 1.7068, 0.0000, 0.032000, 0.320, 0.003469,  1700.1, 0.00117, 1/2000),
    15 : (125.6422, 1.1268, 0.0098, 0.050200, 0.760, 0.003764, 11810.4, 0.00137, 1/2000),
    16 : (324.6287, 2.1304, 0.0100, 0.057400, 0.760, 0.002032, 22656.5, 0.00110, 1/2000),
    22 : (154.4388, 0.9325, 0.0066, 0.070000, 0.620, 0.002893,  9641.1, 0.00055, 1/2000),
    23 : ( 26.4653, 0.5250, 0.0034, 0.053400, 0.800, 0.002031,    61.6, 0.00137, 1/2000),
    24 : ( 71.6923, 1.0750, 0.0059, 0.043900, 1.060, 0.003733,  2706.4, 0.00188, 1/2000),
    35 : ( 49.7384, 0.6888, 0.0088, 0.036000, 1.060, 0.004105,  4534.7, 0.00196, 1/2000),
    38 : ( 24.3404, 0.7000, 0.0040, 0.025321, 0.412, 0.004153,  3062.6, 0.00171, 1/1250),
    41 : ( 58.8110, 0.9250, 0.0033, 0.025321, 0.422, 0.002749, 10013.1, 0.00171, 1/1250),
    42 : ( 21.8254, 0.4625, 0.0019, 0.026194, 0.442, 0.001241,  1090.8, 0.00171, 1/1250),
    43 : (340.5081, 4.2975, 0.0043, 0.025321, 0.448, 0.002043, 19767.6, 0.00171, 1/1250),
    44 : ( 24.0977, 0.7300, 0.0054, 0.031651, 0.316, 0.003485, 37596.3, 0.00033, 1/1250),
    45 : (  3.4375, 0.1375, 0.0069, 0.033027, 0.320, 0.002397, 10421.2, 0.00016, 1/1250),
    47 : (  8.7813, 0.3513, 0.0026, 0.029000, 0.358, 0.003257,  1369.0, 0.00171, 1/1250),
    48 : ( 35.6250, 1.4250, 0.0063, 0.023019, 0.496, 0.003076,  7046.4, 0.00171, 1/1250),
    49 : ( 20.0000, 0.8000, 0.0046, 0.034529, 0.304, 0.003744,   823.3, 0.00171, 1/1250),
    50 : (  8.1250, 0.3250, 0.0000, 0.033027, 0.320, 0.004033,  2118.5, 0.00171, 1/1250),
    51 : ( 15.0000, 0.6000, 0.0071, 0.036173, 0.294, 0.004315,   570.4, 0.00171, 1/1250),
    52 : ( 49.2200, 1.6075, 0.0047, 0.036173, 0.304, 0.001716,  4025.6, 0.00171, 1/1250),
    53 : ( 69.4565, 1.1625, 0.0028, 0.031651, 0.336, 0.002700,  9819.5, 0.00171, 1/1250)}
data = {i : {k : v for k, v in zip(params, raw_data[i])} for i in raw_data.keys()}

# Set the ring we are analyzing
ring = 15
max_failure_probability = data[ring]["max_Pf"]

# Compute the investment cost to increase the dike height
def exponential_investment_cost(u,     # increase in dike height
                                h0,    # original height of the dike
                                c,     # constant from Table 1
                                b,     # constant from Table 1
                                lam):  # constant from Table 1
    if u == 0:
        return 0
    else:
        return (c + b*u)*math.exp(lam*(h0+u))

# The Python function implementing the model
def eijgenraam(Xs,                          # list of dike heightenings
               Ts,                          # time of dike heightenings
               T = 300,                     # planning horizon
               P0 = data[ring]["P0"],       # constant from Table 1
               V0 = data[ring]["V0"],       # constant from Table 1
               alpha = data[ring]["alpha"], # constant from Table 1
               delta = 0.04,                # discount rate, mentioned in Section 2.2
               eta = data[ring]["eta"],     # constant from Table 1
               gamma = 0.035,               # paper says this is taken from government report, but no indication of actual value
               rho = 0.015,                 # risk-free rate, mentioned in Section 2.2
               zeta = data[ring]["zeta"],   # constant from Table 1
               c = data[ring]["c"],         # constant from Table 1
               b = data[ring]["b"],         # constant from Table 1
               lam = data[ring]["lam"]):    # constant from Table 1
    Ts = [int(Ts[i] + sum(Ts[:i])) for i in range(len(Ts)) if Ts[i] + sum(Ts[:i]) < T]
    Xs = Xs[:len(Ts)]

    if len(Ts) == 0:
        Ts = [0]
        Xs = [0]

    if Ts[0] > 0:
        Ts.insert(0, 0)
        Xs.insert(0, 0)

    S0 = P0*V0
    beta = alpha*eta + gamma - rho
    theta = alpha - zeta

    # calculate investment
    investment = 0

    for i in range(len(Xs)):
        step_cost = exponential_investment_cost(Xs[i], 0 if i==0 else sum(Xs[:i]), c, b, lam)
        step_discount = math.exp(-delta*Ts[i])
        investment += step_cost * step_discount

    # calculate expected losses
    losses = 0

    for i in range(len(Xs)-1):
        losses += math.exp(-theta*sum(Xs[:(i+1)]))*(math.exp((beta - delta)*Ts[i+1]) - math.exp((beta - delta)*Ts[i]))

    if Ts[-1] < T:
        losses += math.exp(-theta*sum(Xs))*(math.exp((beta - delta)*T) - math.exp((beta - delta)*Ts[-1]))

    losses = losses * S0 / (beta - delta)

    # salvage term
    losses += S0*math.exp(beta*T)*math.exp(-theta*sum(Xs))*math.exp(-delta*T) / delta

    def find_height(t):
        if t < Ts[0]:
            return 0
        elif t > Ts[-1]:
            return sum(Xs)
        else:
            return sum(Xs[:bisect.bisect_right(Ts, t)])

    failure_probability = [P0*np.exp(alpha*eta*t)*np.exp(-alpha*find_height(t)) for t in range(T+1)]
    total_failure = 1 - functools.reduce(operator.mul, [1 - p for p in failure_probability], 1)
    mean_failure = sum(failure_probability) / (T+1)
    max_failure = max(failure_probability)

    return (investment, losses, investment+losses, total_failure, mean_failure, max_failure)

# Generate a plot showing the dike heightenings and failure probability over time
def plot_details(
               Xs,                          # list of dike heightenings
               Ts,                          # time of dike heightenings
               T = 300,                     # planning horizon
               P0 = data[ring]["P0"],       # constant from Table 1
               alpha = data[ring]["alpha"], # constant from Table 1
               eta = data[ring]["eta"],     # constant from Table 1
               plot_args = {}):   
    Ts = [int(Ts[i] + sum(Ts[:i])) for i in range(len(Ts)) if Ts[i] + sum(Ts[:i]) < T]
    Xs = Xs[:len(Ts)]

    if len(Ts) == 0:
        Ts = [0]
        Xs = [0]

    # convert inputs to numpy arrays
    P0 = np.asarray([P0]) if isinstance(P0, numbers.Number) else np.asarray(P0)
    alpha = np.asarray([alpha]) if isinstance(alpha, numbers.Number) else np.asarray(alpha)
    eta = np.asarray([eta]) if isinstance(eta, numbers.Number) else np.asarray(eta)
    n = max([x.shape[0] for x in [P0, alpha, eta]])

    # compute the failure probability
    def find_height(t):
        if t < Ts[0]:
            return 0
        elif t > Ts[-1]:
            return sum(Xs)
        else:
            return sum(Xs[:bisect.bisect_right(Ts, t)])

    failure_probability = np.empty((n, T+1))

    for t in range(T+1):
        failure_probability[:,t] = P0*np.exp(alpha*eta*t)*np.exp(-alpha*find_height(t))

    # generate the plot
    fig = plt.figure()

    for i in range(n):
        plt.plot(range(T+1), failure_probability[i,:], 'b-', **plot_args)

    plt.plot(range(T+1), [max_failure_probability]*(T+1), 'r--')

    for i in range(len(Ts)):
        if Ts[i] == 0:
            plt.text(0, np.max(failure_probability[:,Ts[0]])/2, str(round(Xs[i], 1)) + " cm", ha='left', va='center')
        else:
            plt.text(Ts[i], (np.max(failure_probability[:,Ts[i]-1])+np.max(failure_probability[:,Ts[i]]))/2, str(round(Xs[i], 1)) + " cm", ha='left', va='center')

    if n == 1:
        plt.legend(["Failure Probability", "Current Safety Standard"])
    else:
        plt.ylim([0, 0.001])

    plt.xlabel("Time (years)")
    plt.ylabel("Failure Probability")
    plt.show()

##==============================================================================
## Create the model in Rhodium
##------------------------------------------------------------------------------
model = Model(eijgenraam)

model.parameters = [Parameter("Xs"),
                    Parameter("Ts"),
                    Parameter("T"),
                    Parameter("P0"),
                    Parameter("V0"),
                    Parameter("alpha"),
                    Parameter("delta"),
                    Parameter("eta"),
                    Parameter("gamma"),
                    Parameter("rho"),
                    Parameter("zeta"),
                    Parameter("c"),
                    Parameter("b"),
                    Parameter("lam")]

model.responses = [Response("TotalInvestment", Response.INFO),
                   Response("TotalLoss", Response.INFO),
                   Response("TotalCost", Response.MINIMIZE),
                   Response("TotalFailureProb", Response.INFO),
                   Response("AvgFailureProb", Response.MINIMIZE),
                   Response("MaxFailureProb", Response.MINIMIZE)]

# Include constraints to keep costs within a reasonable range
model.constraints = [Constraint("TotalInvestment <= 1000.0"),
                     Constraint("TotalLoss <= 50000.0"),
                     Constraint("MaxFailureProb <= " + str(data[ring]["max_Pf"]))]

# Set uncertainties
model.uncertainties = [LogNormalUncertainty("P0", 0.00137, 0.25),
                       NormalUncertainty("alpha", 0.0502, 0.01),
                       LogNormalUncertainty("eta", 0.76, 0.1)]

# Each height/time pair defines the dike height increase, Xs[i], after waiting
# some number of years, Ts[i], from the previous dike heightening.  If
# sum(Ts[0..i]) exceeds the planning horizon, T, then any remaining heightenings
# are ignored.
model.levers = [RealLever("Xs", 0, 500, length=6),
                RealLever("Ts", 0, 300, length=6)]


##==============================================================================
## Setup cache
##------------------------------------------------------------------------------
setup_cache(file="eijgenraam_ring%d.cache" % ring)

##==============================================================================
## Optimize the model (caching result to avoid recomputing each time)
##------------------------------------------------------------------------------
policies = cache("policies_mordm", lambda: optimize(model, "NSGAII", 10000))

##==============================================================================
## Export the data (Excel, CSV, JSON, pickle)
##------------------------------------------------------------------------------
if promptToRun("Save policies to Excel (eijgenraam_mordm.xlsx)?"):
    policies.save("eijgenraam_mordm.xlsx")

##==============================================================================
## Display all the data in a pretty format
##------------------------------------------------------------------------------
if promptToRun("Print all policies?"):
    print(policies)

##==============================================================================
## Query and manipulate the data
##------------------------------------------------------------------------------
subset = policies[:5]                        # get first five entries in data
values = policies["MaxFailureProb"]          # pull out a single key (returned as a list
result = policies.find("TotalCost <= 600.0") # search for data matching a query
policy = policies.find_min("TotalCost")      # find min/max policy
df = policies.as_dataframe()                 # convert to Pandas data frame
arr = policies.as_array()                    # convert to Numpy array

# apply a function; if we include an assignment, a new field in the dataset
# will be created
policies.apply("PercentageInvestment = TotalInvestment / TotalCost")

##==============================================================================
## Show 3d scatter plot
##------------------------------------------------------------------------------
if promptToRun("Display 3D scatter plot?"):
    scatter3d(model, policies)
    plt.show()

##==============================================================================
## Show 3d scatter plot that can be clicked to show details
##------------------------------------------------------------------------------
if promptToRun("Display 3D scatter plot with interactive points?"):
    scatter3d(model, policies, x="TotalInvestment", y="TotalLoss", z="MaxFailureProb",
              depthshade=False, interactive=False,
              pick_handler=lambda i : plot_details(policies[i]["Xs"], policies[i]["Ts"]))
    plt.show()

##==============================================================================
## Create rotating animation of 3D scatter plot
##------------------------------------------------------------------------------
if promptToRun("Create animated GIF showing the 3D scatter plot?"):
    scatter3d(model, policies, x="TotalInvestment", y="TotalLoss", z="MaxFailureProb",
              depthshade=False)
    animate3d("eijgenraam", steps=360, transform=(1, 0, 0), duration=0.02)

##==============================================================================
## Plot minimum cost, minimum failure probability policies
##------------------------------------------------------------------------------
if promptToRun("Display details for the policy minimizing total cost?"):
    policy = policies.find_min("TotalCost")
    plot_details(policy["Xs"], policy["Ts"])

##==============================================================================
## Test a policy against uncertainties
##------------------------------------------------------------------------------      
SOWs = sample_lhs(model, 1000)

if promptToRun("Display uncertainty SOWs for the policy minimizing total cost?"):
    policy = policies.find_min("TotalCost")
    plot_details(policy["Xs"], policy["Ts"], P0=SOWs["P0"], alpha=SOWs["alpha"], eta=SOWs["eta"], plot_args={"alpha" : 0.02})

##==============================================================================
## Identify the key uncertainties
##------------------------------------------------------------------------------
policy = policies.find_min("TotalCost")
SOW_output = cache("SOW_output", lambda : evaluate(model, overwrite(SOWs, policy)))

threshold = lambda x : "Succeeds" if x["MaxFailureProb"] < max_failure_probability else "Fails"
metric = SOW_output.apply(threshold)

if promptToRun("Show PRIM analysis of key uncertainties?"):
    p = Prim(SOW_output, metric, include=model.uncertainties.keys(), coi="Fails")
    box = p.find_box()
    box.show_tradeoff()
    plt.show()

if promptToRun("Show CART analysis of key uncertainties?"):
    c = Cart(SOW_output, metric, include=model.uncertainties.keys(), min_samples_leaf=50)
    c.print_tree(coi="Fails")
    c.show_tree()
    plt.show()

##==============================================================================
## Sensitivity analysis of uncertainties
##------------------------------------------------------------------------------
if promptToRun("Show sensitivity analysis of max failure prop w.r.t. uncertainties"):
    policy = policies.find_min("TotalCost")
    result = cache("sa", lambda : sa(model, "MaxFailureProb", policy=policy, method="sobol", nsamples=100000))

    print(result)

    result.plot_sobol(threshold=0.0)
    plt.show()

##==============================================================================
## One-at-a-time sensitivity analysis of uncertainties
##------------------------------------------------------------------------------

policy = policies.find_min("TotalCost")

if promptToRun("Show one-at-a-time sensitivity analysis of uncertainties?"):
    oat(model, "MaxFailureProb", policy=policy, nsamples=100)
    plt.show()

if promptToRun("Show regional sensitivity analysis of uncertainties?"):
    threshold = lambda x : "Succeeds" if x["MaxFailureProb"] < max_failure_probability else "Fails"
    regional_sa(model, threshold, policy=policy, nsamples=1000)
    plt.show()

##==============================================================================
## Compute robustness measures
##------------------------------------------------------------------------------
if promptToRun("Show 3D scatter plot with robustness color code?"):
    model.constraints = [Constraint("MaxFailureProb <= " + str(data[ring]["max_Pf"]))]

    SOWs = sample_lhs(model, 100)
    results = cache("robustness", lambda : evaluate_robustness(model, policies, SOWs))
    scatter3d(model, results, x="TotalInvestment", y="TotalLoss", z="MaxFailureProb", c="Satisficing Type 1", depthshade=False, interactive=False)
    plt.show()



from rhodium import *
from rhodium.excel import ExcelModel

# Provide the Excel file
file = os.path.abspath(os.path.join(os.path.dirname(__file__), "LakeProblem.xlsx"))

with ExcelModel(file) as model:
    # The parameter names must match the R arguments exactly
    model.parameters = [Parameter("pollution_limit", sheet="Inputs", cell="A2:A101"),
                        Parameter("b", sheet="Inputs", cell="B2"),
                        Parameter("q", sheet="Inputs", cell="C2"),
                        Parameter("mean", sheet="Inputs", cell="D2"),
                        Parameter("stdev", sheet="Inputs", cell="E2"),
                        Parameter("delta", sheet="Inputs", cell="F2")]

    # List all outputs from the R function, which should return these values either as
    # an unnamed array in the given order (e.g., c(max_P, utility, ...)) or as a named
    # list (e.g., list(max_P=..., utility=..., ...)).
    model.responses = [Response("max_P", Response.MINIMIZE, sheet="Calculations", cell="A8"),
                       Response("utility", Response.MAXIMIZE, sheet="Calculations", cell="A11"),
                       Response("inertia", Response.MAXIMIZE, sheet="Calculations", cell="A14"),
                       Response("reliability", Response.MAXIMIZE, sheet="Calculations", cell="A5")]

    # Specify the levers
    model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

    # Optimize the model using Rhodium
    output = optimize(model, "NSGAII", 10000)
    print(output)


import math
import numpy as np
from scipy.optimize import brentq as root
from rhodium import *

# Configure the lake model
def lake_problem(pollution_limit,
         b = 0.42,        # decay rate for P in lake (0.42 = irreversible)
         q = 2.0,         # recycling exponent
         mean = 0.02,     # mean of natural inflows
         stdev = 0.001,   # standard deviation of natural inflows
         alpha = 0.4,     # utility from pollution
         delta = 0.98,    # future utility discount rate
         nsamples = 100): # monte carlo sampling of natural inflows
    Pcrit = root(lambda x: x**q/(1+x**q) - b*x, 0.01, 1.5)
    nvars = len(pollution_limit)
    X = np.zeros((nvars,))
    average_daily_P = np.zeros((nvars,))
    decisions = np.array(pollution_limit)
    reliability = 0.0

    for _ in range(nsamples):
        X[0] = 0.0

        natural_inflows = np.random.lognormal(
                math.log(mean**2 / math.sqrt(stdev**2 + mean**2)),
                math.sqrt(math.log(1.0 + stdev**2 / mean**2)),
                size = nvars)

        for t in range(1,nvars):
            X[t] = (1-b)*X[t-1] + X[t-1]**q/(1+X[t-1]**q) + decisions[t-1] + natural_inflows[t-1]
            average_daily_P[t] += X[t]/float(nsamples)

        reliability += np.sum(X < Pcrit)/float(nsamples*nvars)

    max_P = np.max(average_daily_P)
    utility = np.sum(alpha*decisions*np.power(delta,np.arange(nvars)))
    intertia = np.sum(np.diff(decisions) > -0.02)/float(nvars-1)

    return (max_P, utility, intertia, reliability)

model = Model(lake_problem)

model.parameters = [Parameter("pollution_limit"),
                    Parameter("b"),
                    Parameter("q"),
                    Parameter("mean"),
                    Parameter("stdev"),
                    Parameter("delta")]

model.responses = [Response("max_P", Response.MINIMIZE),
                   Response("utility", Response.MAXIMIZE),
                   Response("inertia", Response.MAXIMIZE),
                   Response("reliability", Response.MAXIMIZE)]

model.levers = [RealLever("pollution_limit", 0.0, 0.1, length=100)]

model.uncertainties = [UniformUncertainty("b", 0.1, 0.45),
                       UniformUncertainty("q", 2.0, 4.5),
                       UniformUncertainty("mean", 0.01, 0.05),
                       UniformUncertainty("stdev", 0.001, 0.005),
                       UniformUncertainty("delta", 0.93, 0.99)]

# Rhodium performs sensitivity analysis on the uncertainties with respect to a response.
# For example, here we are interested in the sensitivity of the reliability response to
# the uncertain parameters.  Here, we'll perform the analysis on a single policy.
#
# It's important to understand the underlying assumptions of each method.  Some methods
# may require fewer or more samples to produce accurate results.  Additionally, different
# methods report sensitivities in different ways; typically we recommend using the ranking
# of sensitivities rather than the relative value to measure the importance of different
# uncertainties.
#
# Rhodium currently supports:
#
#    "fast" - Fourier Amplitude Sensitivity Test,
#    "morris" - Method of Morris,
#    "delta" - Borgonovo's Delta Moment-Independent Method,
#    "sobol" - Sobol Sensitivity Analysis,
#    "ff" - Fractional Factorial Sensitivity Analysis, and
#    "dgsm" - Derivative-based Global Sensitivity Measure.
#
# Several of these methods are demonstrated below.
#
# In addition to the text output, you can call plot() on the results object to visualize
# the data in a plot.  Sobol's method supports a special radial plot accessible by calling
# plot_sobol().

policy = { "pollution_limit" : [0.075]*100 }

fast_results = sa(model, "reliability", policy=policy, method="fast", nsamples=1000)
print("Fourier Amplitude Sensitivity Test (FAST)")
print("------------------------------------------------------------")
print(fast_results)
print()

delta_results = sa(model, "reliability", policy=policy, method="delta", nsamples=1000)
print("Borgonovo's Delta Moment-Independent Measure")
print("------------------------------------------------------------")
print(delta_results)
print()

sobol_results = sa(model, "reliability", policy=policy, method="sobol", nsamples=10000)
print("Sobol Sensitivity Analysis")
print("------------------------------------------------------------")
print(sobol_results)
print()

morris_results = sa(model, "reliability", policy=policy, method="morris", nsamples=1000, num_levels=4, grid_jump=2)
print("Method of Morris")
print("------------------------------------------------------------")
print(morris_results)
print()
